# -*- coding:utf-8 -*-
from sqlalchemy.ext.declarative import declarative_base
import sqlalchemy as sa
Base = declarative_base()


def _getTarget():
    from alchemyjsonschema import SchemaFactory
    return SchemaFactory


def _makeOne(*args, **kwargs):
    return _getTarget()(*args, **kwargs)


def test_detect__nullable_is_True__not_required():
    from alchemyjsonschema import SingleModelWalker

    class Model0(Base):
        __tablename__ = "Model0"
        pk = sa.Column(sa.Integer, primary_key=True, doc="primary key")
        created_at = sa.Column(sa.DateTime, nullable=True)

    target = _makeOne(SingleModelWalker)
    walker = target.walker(Model0)
    result = target._detect_required(walker)
    assert result == ["pk"]


def test_detect__nullable_is_False__required():
    from alchemyjsonschema import SingleModelWalker

    class Model1(Base):
        __tablename__ = "Model1"
        pk = sa.Column(sa.Integer, primary_key=True, doc="primary key")
        created_at = sa.Column(sa.DateTime, nullable=False)

    target = _makeOne(SingleModelWalker)
    walker = target.walker(Model1)
    result = target._detect_required(walker)
    assert result == ["pk", "created_at"]


def test_detect__nullable_is_False__but_default_is_exists__not_required():
    from alchemyjsonschema import SingleModelWalker
    from datetime import datetime

    class Model2(Base):
        __tablename__ = "Model2"
        pk = sa.Column(sa.Integer, primary_key=True, doc="primary key")
        created_at = sa.Column(sa.DateTime, nullable=False, default=datetime.now)

    target = _makeOne(SingleModelWalker)
    walker = target.walker(Model2)
    result = target._detect_required(walker)
    assert result == ["pk"]



# -*- coding:utf-8 -*-
import sys
import pkg_resources
import json
import argparse


from . import (
    SchemaFactory,
    AlsoChildrenWalker,
    OneModelOnlyWalker,
    SingleModelWalker,
    HandControlledWalkerFactory,
    RelationDesicion,
    ComfortableDesicion
)


def import_symbol(symbol):  # todo cache
    return pkg_resources.EntryPoint.parse("x=%s" % symbol).load(False)


def err(x):
    sys.stderr.write(x)
    sys.stderr.write("\n")
    sys.stderr.flush()


def detect_walker(x):
    if x == "structual":
        return AlsoChildrenWalker
    elif x == "noforeignkey":
        return OneModelOnlyWalker
    elif x == "foreignkey":
        return SingleModelWalker
    elif x == "control":
        return HandControlledWalkerFactory
    else:
        raise Exception(x)


def detect_decision(x):
    if x == "default":
        return RelationDesicion()
    elif x == "comfortable":
        return ComfortableDesicion()
    else:
        raise Exception(x)


def run(model, walker, depth=None, relation_decision=None):
    make_schema = SchemaFactory(walker, relation_decision=relation_decision)
    schema = make_schema(model, depth=depth)
    print(json.dumps(schema, indent=2, ensure_ascii=False))


def main(sys_args=sys.argv[1:]):
    parser = argparse.ArgumentParser()
    parser.add_argument("model")
    parser.add_argument("--walker", choices=["noforeignkey", "foreignkey", "structual", "control"], default="structual")
    parser.add_argument("--decision", choices=["default", "comfortable"], default="default")
    parser.add_argument("--depth", default=None, type=int)
    parser.add_argument("--decision-relationship", default="")
    parser.add_argument("--decision-foreignkey", default="")
    args = parser.parse_args(sys_args)
    walker = detect_walker(args.walker)
    model = import_symbol(args.model)
    if walker == HandControlledWalkerFactory:
        decisions = {k.strip(): "relationship" for k in args.decision_relationship.split(" ")}
        decisions.update({k.strip(): "foreignkey" for k in args.decision_foreignkey.split(" ")})
        walker = walker(decisions)
    return run(model, walker, depth=args.depth, relation_decision=detect_decision(args.decision))



# -*- coding:utf-8 -*-
def _getTarget():
    from alchemyjsonschema import SchemaFactory
    return SchemaFactory


def _makeOne(*args, **kwargs):
    from alchemyjsonschema import (
        SingleModelWalker,
        DefaultClassfier
    )
    return _getTarget()(SingleModelWalker, DefaultClassfier)


# definition
import sqlalchemy as sa
import sqlalchemy.orm as orm
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()


class Group(Base):
    """model for test"""
    __tablename__ = "Group"

    pk = sa.Column(sa.Integer, primary_key=True, doc="primary key")
    name = sa.Column(sa.String(255), default="", nullable=False)
    color = sa.Column(sa.Enum("red", "green", "yellow", "blue"))


class User(Base):
    __tablename__ = "User"

    pk = sa.Column(sa.Integer, primary_key=True, doc="primary key")
    name = sa.Column(sa.String(255), default="", nullable=True)
    group_id = sa.Column(sa.Integer, sa.ForeignKey(Group.pk), nullable=False)
    group = orm.relationship(Group, uselist=False, backref="users")


def test_it_create_schema__and__valid_params__sucess():
    from jsonschema import validate

    target = _makeOne()
    schema = target(Group, excludes=["pk", "users.pk"])
    data = {"name": "ravenclaw", "color": "blue", "users": [{"name": "foo"}, {"name": "bar"}]}

    validate(data, schema)


def test_it_creat_schema__and__invalid_params__failure():
    import pytest
    from jsonschema import validate
    from jsonschema.exceptions import ValidationError

    target = _makeOne()
    schema = target(Group, excludes=["pk", "uesrs.pk"])
    data = {"name": "blackmage", "color": "black", "users": [{"name": "foo"}, {"name": "bar"}]}

    with pytest.raises(ValidationError):
        validate(data, schema)


def test_it2_create_schema__and__valid_params__success():
    from jsonschema import validate

    target = _makeOne()
    schema = target(User, excludes=["pk", "group_id"])
    data = {"name": "foo", "group": {"name": "ravenclaw", "color": "blue", "pk": 1}}
    validate(data, schema)


def test_it_jsonify_data__that_is_valid_params():
    from alchemyjsonschema.dictify import jsonify
    from jsonschema import validate

    target = _makeOne()
    schema = target(User, excludes=["pk", "group_id"])

    user = User(name="foo", group=Group(name="ravenclaw", color="blue", pk=1))
    jsondict = jsonify(user, schema)
    validate(jsondict, schema)



# -*- coding:utf-8 -*-
def _callFUT(*args, **kwargs):
    from alchemyjsonschema.dictify import apply_changes
    return apply_changes(*args, **kwargs)


def test_it_update_parent__onlyone():
    from alchemyjsonschema.tests import models as m
    from alchemyjsonschema import SchemaFactory, OneModelOnlyWalker
    from alchemyjsonschema.dictify import ModelLookup

    factory = SchemaFactory(OneModelOnlyWalker)
    a_schema = factory(m.A0)
    modellookup = ModelLookup(m)

    a0 = m.A0(pk=1, name="a0", children=[m.A1(pk=1, name="a00"), m.A1(pk=2, name=("a01"))])

    assert a0.name == "a0"
    params = {"pk": 1, "name": "updated"}
    _callFUT(a0, params, a_schema, modellookup)
    assert a0.name == "updated"
    assert len(a0.children) == 2


def test_it_update_parent__full():
    from alchemyjsonschema.tests import models as m
    from alchemyjsonschema import SchemaFactory, AlsoChildrenWalker
    from alchemyjsonschema.dictify import ModelLookup

    factory = SchemaFactory(AlsoChildrenWalker)
    a_schema = factory(m.A0)
    modellookup = ModelLookup(m)

    a0 = m.A0(pk=1, name="a0", children=[m.A1(pk=1, name="a00"), m.A1(pk=2, name=("a01"))])

    assert a0.name == "a0"
    params = {"pk": 1, "name": "updated", "children": [{"name": "a00", "pk": 1}, {"name": "a01", "pk": 2}]}
    _callFUT(a0, params, a_schema, modellookup)
    assert a0.name == "updated"


def test_it_create_child():
    from alchemyjsonschema.tests import models as m
    from alchemyjsonschema import SchemaFactory, AlsoChildrenWalker
    from alchemyjsonschema.dictify import ModelLookup

    factory = SchemaFactory(AlsoChildrenWalker)
    a_schema = factory(m.A0)
    modellookup = ModelLookup(m)

    a0 = m.A0(pk=1, name="a0", children=[m.A1(pk=1, name="a00"), m.A1(pk=2, name=("a01"))])

    assert len(a0.children) == 2
    params = {"pk": 1, "name": "a0", "children": [{"name": "new"}, {"name": "a00", "pk": 1}, {"name": "update", "pk": 2}]}
    _callFUT(a0, params, a_schema, modellookup)
    assert len(a0.children) == 3


def test_it_update_child():
    from alchemyjsonschema.tests import models as m
    from alchemyjsonschema import SchemaFactory, AlsoChildrenWalker
    from alchemyjsonschema.dictify import ModelLookup

    factory = SchemaFactory(AlsoChildrenWalker)
    a_schema = factory(m.A0)
    modellookup = ModelLookup(m)

    a0 = m.A0(pk=1, name="a0", children=[m.A1(pk=1, name="a00"), m.A1(pk=2, name=("a01"))])

    assert len(a0.children) == 2
    params = {"pk": 1, "name": "a0", "children": [{"name": "a00", "pk": 1}, {"name": "update", "pk": 2}]}
    _callFUT(a0, params, a_schema, modellookup)
    assert len(a0.children) == 2
    assert a0.children[1].name == "update"


def test_it_delete_child():
    from alchemyjsonschema.tests import models as m
    from alchemyjsonschema import SchemaFactory, AlsoChildrenWalker
    from alchemyjsonschema.dictify import ModelLookup

    factory = SchemaFactory(AlsoChildrenWalker)
    a_schema = factory(m.A0)
    modellookup = ModelLookup(m)

    a0 = m.A0(pk=1, name="a0", children=[m.A1(pk=1, name="a00"), m.A1(pk=2, name=("a01"))])

    assert len(a0.children) == 2
    params = {"pk": 1, "name": "a0", "children": [{"name": "a00", "pk": 1}]}
    _callFUT(a0, params, a_schema, modellookup)
    assert len(a0.children) == 1


def test_it_reverse_update():
    from alchemyjsonschema.tests import models as m
    from alchemyjsonschema import SchemaFactory, AlsoChildrenWalker
    from alchemyjsonschema.dictify import ModelLookup

    factory = SchemaFactory(AlsoChildrenWalker)
    a_schema = factory(m.A1)
    modellookup = ModelLookup(m)

    a0 = m.A0(pk=1, name="a0", children=[m.A1(pk=1, name="a00"), m.A1(pk=2, name=("a01"))])

    assert a0.name == "a0"
    params = {"pk": 1, "name": "a00", "parent": {"name": "updated", "pk": 1}}
    _callFUT(a0.children[0], params, a_schema, modellookup)
    assert a0.name == "updated"


# def test_it_complex():
#     pass



# -*- coding:utf-8 -*-
"""
jsondict <-> dict <-> model object
   \______________________/
"""


def _datetime(*args):
    import pytz
    from datetime import datetime
    args = list(args)
    args.append(pytz.utc)
    return datetime(*args)


def _getTarget():
    from alchemyjsonschema.mapping import Draft4MappingFactory
    return Draft4MappingFactory


def _makeOne(schema_factory, model, *args, **kwargs):
    import alchemyjsonschema.tests.models as models
    module = models
    mapping_factory = _getTarget()(schema_factory, module, *args, **kwargs)
    return mapping_factory(model)


def test_it__dict_from_model_object():
    from alchemyjsonschema import AlsoChildrenWalker, SchemaFactory
    from .models import Group, User

    schema_factory = SchemaFactory(AlsoChildrenWalker)
    target = _makeOne(schema_factory, Group)

    group = Group(name="ravenclaw", color="blue", created_at=_datetime(2000, 1, 1, 10, 0, 0, 0))
    group.users = [User(name="foo", created_at=_datetime(2000, 1, 1, 10, 0, 0, 0))]

    group_dict = target.dict_from_object(group)
    assert group_dict == {'color': 'blue',
                          'users': [{'created_at': _datetime(2000, 1, 1, 10, 0, 0, 0), 'pk': None, 'name': 'foo'}],
                          'created_at': _datetime(2000, 1, 1, 10, 0, 0, 0),
                          'pk': None,
                          'name': 'ravenclaw'}


def test_it__jsondict_from_model():
    from alchemyjsonschema import AlsoChildrenWalker, SchemaFactory
    from .models import Group, User

    schema_factory = SchemaFactory(AlsoChildrenWalker)
    target = _makeOne(schema_factory, Group)

    group = Group(name="ravenclaw", color="blue", created_at=_datetime(2000, 1, 1, 10, 0, 0, 0))
    group.users = [User(name="foo", created_at=_datetime(2000, 1, 1, 10, 0, 0, 0))]

    jsondict = target.jsondict_from_object(group)

    import json
    assert json.dumps(jsondict)

    assert jsondict == {'color': 'blue',
                        'name': 'ravenclaw',
                        'users': [{'name': 'foo', 'pk': None, 'created_at': '2000-01-01T10:00:00+00:00'}],
                        'pk': None,
                        'created_at': '2000-01-01T10:00:00+00:00'}


def test_it__validate__jsondict():
    from alchemyjsonschema import AlsoChildrenWalker, SchemaFactory
    from .models import Group

    schema_factory = SchemaFactory(AlsoChildrenWalker)
    target = _makeOne(schema_factory, Group)

    jsondict = {'color': 'blue',
                'name': 'ravenclaw',
                'users': [{'name': 'foo', 'pk': 1, 'created_at': '2000-01-01T10:00:00+00:00'}],
                'pk': 1,
                'created_at': '2000-01-01T10:00:00+00:00'}

    target.validate_jsondict(jsondict)


def test_it__dict_from_jsondict():
    from alchemyjsonschema import AlsoChildrenWalker, SchemaFactory
    from .models import Group

    schema_factory = SchemaFactory(AlsoChildrenWalker)
    target = _makeOne(schema_factory, Group)

    jsondict = {'color': 'blue',
                'name': 'ravenclaw',
                'users': [{'name': 'foo', 'pk': 10, 'created_at': '2000-01-01T10:00:00+00:00'}],
                'pk': None,
                'created_at': '2000-01-01T10:00:00+00:00'}

    group_dict = target.dict_from_jsondict(jsondict)

    assert group_dict == {'color': 'blue',
                          'users': [{'created_at': _datetime(2000, 1, 1, 10, 0, 0, 0), 'pk': 10, 'name': 'foo'}],
                          'created_at': _datetime(2000, 1, 1, 10, 0, 0, 0),
                          'pk': None,
                          'name': 'ravenclaw'}


def test_it__object_from_dict():
    from alchemyjsonschema import AlsoChildrenWalker, SchemaFactory
    from .models import Group, User

    schema_factory = SchemaFactory(AlsoChildrenWalker)
    target = _makeOne(schema_factory, Group)

    group_dict = {'color': 'blue',
                  'users': [{'created_at': _datetime(2000, 1, 1, 10, 0, 0, 0), 'pk': None, 'name': 'foo'}],
                  'created_at': _datetime(2000, 1, 1, 10, 0, 0, 0),
                  'pk': None,
                  'name': 'ravenclaw'}

    group = target.object_from_dict(group_dict, strict=False)

    assert isinstance(group, Group)
    assert group.color == "blue"
    assert group.name == "ravenclaw"
    assert group.pk is None
    assert group.created_at == _datetime(2000, 1, 1, 10, 0, 0, 0)

    assert (len(group.users) == 1) and (isinstance(group.users[0], User))
    assert group.users[0].name == "foo"
    assert group.users[0].pk is None
    assert group.users[0].created_at == _datetime(2000, 1, 1, 10, 0, 0, 0)




# -*- coding:utf-8 -*-
def _getTarget():
    from alchemyjsonschema import SchemaFactory
    return SchemaFactory


def _makeOne(walker, *args, **kwargs):
    from alchemyjsonschema import DefaultClassfier
    return _getTarget()(walker, DefaultClassfier, *args, **kwargs)


# definition
import sqlalchemy as sa
import sqlalchemy.orm as orm
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()


class Group(Base):
    """model for test"""
    __tablename__ = "Group"

    pk = sa.Column(sa.Integer, primary_key=True, doc="primary key")
    name = sa.Column(sa.String(255), default="", nullable=False)


class User(Base):
    __tablename__ = "User"

    pk = sa.Column(sa.Integer, primary_key=True, doc="primary key")
    name = sa.Column(sa.String(255), default="", nullable=True)
    group_id = sa.Column(sa.Integer, sa.ForeignKey(Group.pk), nullable=False)
    group = orm.relationship(Group, uselist=False, backref="users")


def test_properties__default__includes__foreign_keys():
    from alchemyjsonschema import SingleModelWalker
    target = _makeOne(SingleModelWalker)
    result = target(User)

    assert "properties" in result
    assert list(sorted(result["properties"].keys())) == ["group_id", "name", "pk"]


def test_properties__include_OnetoMany_relation():
    from alchemyjsonschema import AlsoChildrenWalker, RelationDesicion
    target = _makeOne(AlsoChildrenWalker, relation_decision=RelationDesicion())
    result = target(User)

    assert "required" in result
    assert list(sorted(result["properties"])) == ["group", "name", "pk"]
    assert result["properties"]["group"] == {"$ref": "#/definitions/Group"}


def test_properties__include_OnetoMany_relation2():
    from alchemyjsonschema import AlsoChildrenWalker, ComfortableDesicion
    target = _makeOne(AlsoChildrenWalker, relation_decision=ComfortableDesicion())
    result = target(User)

    assert "required" in result
    assert list(sorted(result["properties"])) == ["group_id", "name", "pk"]
    assert result["properties"]["group_id"] == {'type': 'integer', "relation": "group"}


def test_properties__include_ManytoOne_backref():
    from alchemyjsonschema import AlsoChildrenWalker
    target = _makeOne(AlsoChildrenWalker)
    result = target(Group)

    assert "required" in result
    assert list(sorted(result["properties"])) == ["name", "pk", "users"]
    assert result["properties"]["users"] == {"type": "array", "items": {"$ref": "#/definitions/User"}}
    assert result["definitions"]["User"] == {"type": "object", 'required': ['pk'],
                                             "properties": {'name': {'maxLength': 255, 'type': 'string'},
                                                            'pk': {'description': 'primary key', 'type': 'integer'}}}


def test_properties__include_ManytoOne_backref__bidirectional_is_true():
    from alchemyjsonschema import AlsoChildrenWalker, ChildFactory
    target = _makeOne(AlsoChildrenWalker, child_factory=ChildFactory(".", bidirectional=True))
    result = target(Group)

    assert "required" in result
    assert list(sorted(result["properties"])) == ["name", "pk", "users"]
    assert result["properties"]["users"] == {"type": "array", "items": {"$ref": "#/definitions/User"}}
    assert result["definitions"]["User"] == {"type": "object", 'required': ['pk'],
                                             "properties": {'name': {'maxLength': 255, 'type': 'string'},
                                                            'group': {'$ref': '#/definitions/Group'},
                                                            'pk': {'description': 'primary key', 'type': 'integer'}}}


# depth
class A0(Base):
    __tablename__ = "A0"

    pk = sa.Column(sa.Integer, primary_key=True, doc="primary key")


class A1(Base):
    __tablename__ = "A1"

    pk = sa.Column(sa.Integer, primary_key=True, doc="primary key1")
    parent_id = sa.Column(sa.Integer, sa.ForeignKey(A0.pk), nullable=False)
    parent = orm.relationship(A0, uselist=False, backref="children")


class A2(Base):
    __tablename__ = "A2"

    pk = sa.Column(sa.Integer, primary_key=True, doc="primary key2")
    parent_id = sa.Column(sa.Integer, sa.ForeignKey(A1.pk), nullable=False)
    parent = orm.relationship(A1, uselist=False, backref="children")


class A3(Base):
    __tablename__ = "A3"

    pk = sa.Column(sa.Integer, primary_key=True, doc="primary key3")
    parent_id = sa.Column(sa.Integer, sa.ForeignKey(A2.pk), nullable=False)
    parent = orm.relationship(A2, uselist=False, backref="children")


class A4(Base):
    __tablename__ = "A4"

    pk = sa.Column(sa.Integer, primary_key=True, doc="primary key4")
    parent_id = sa.Column(sa.Integer, sa.ForeignKey(A3.pk), nullable=False)
    parent = orm.relationship(A3, uselist=False, backref="children")


class A5(Base):
    __tablename__ = "A5"

    pk = sa.Column(sa.Integer, primary_key=True, doc="primary key5")
    parent_id = sa.Column(sa.Integer, sa.ForeignKey(A4.pk), nullable=False)
    parent = orm.relationship(A4, uselist=False, backref="children")


def test_properties__default_depth_is__traverse_all_chlidren():
    from alchemyjsonschema import AlsoChildrenWalker
    from alchemyjsonschema.dictify import get_reference
    target = _makeOne(AlsoChildrenWalker)
    result = target(A0)

    assert "required" in result
    assert list(sorted(result["properties"])) == ["children", "pk"]
    children0 = get_reference(result["properties"]["children"]["items"], result)
    children1 = get_reference(children0["properties"]["children"]["items"], result)
    children2 = get_reference(children1["properties"]["children"]["items"], result)
    children3 = get_reference(children2["properties"]["children"]["items"], result)
    children4 = get_reference(children3["properties"]["children"]["items"], result)
    assert children4["properties"]["pk"]["description"] == "primary key5"


def test_properties__default_depth_is__2__traverse_depth2():
    from alchemyjsonschema import AlsoChildrenWalker
    from alchemyjsonschema.dictify import get_reference
    target = _makeOne(AlsoChildrenWalker)
    result = target(A0, depth=2)

    assert "required" in result
    assert list(sorted(result["properties"])) == ["children", "pk"]
    children0 = get_reference(result["properties"]["children"]["items"], result)
    assert children0["properties"]["pk"]["description"] == "primary key1"


def test_properties__default_depth_is__3__traverse_depth3():
    from alchemyjsonschema import AlsoChildrenWalker
    from alchemyjsonschema.dictify import get_reference
    target = _makeOne(AlsoChildrenWalker)
    result = target(A0, depth=3)

    assert "required" in result
    assert list(sorted(result["properties"])) == ["children", "pk"]
    children0 = get_reference(result["properties"]["children"]["items"], result)
    children1 = get_reference(children0["properties"]["children"]["items"], result)
    assert children1["properties"]["pk"]["description"] == "primary key2"


# regression
# X.y -> Y.z -> Z.y -> Y.z -> Z.y

class Y(Base):
    __tablename__ = "y"
    id = sa.Column(sa.Integer, primary_key=True, doc="primary key")
    z_id = sa.Column(sa.Integer, sa.ForeignKey('z.id'))
    zs = orm.relationship("Z", foreign_keys=[z_id])


class X(Base):
    __tablename__ = "x"
    id = sa.Column(sa.Integer, primary_key=True, doc="primary key")
    y_id = sa.Column(sa.Integer, sa.ForeignKey('y.id'))
    ys = orm.relationship(Y, foreign_keys=[y_id])


class Z(Base):
    __tablename__ = "z"
    id = sa.Column(sa.Integer, primary_key=True, doc="primary key")
    y_id = sa.Column(sa.Integer, sa.ForeignKey('y.id'))
    ys = orm.relationship(Y, foreign_keys=[y_id])


def test_properties__infinite_loop():
    from alchemyjsonschema import AlsoChildrenWalker, RelationDesicion
    from alchemyjsonschema.dictify import get_reference
    target = _makeOne(AlsoChildrenWalker, relation_decision=RelationDesicion())
    result = target(X)
    ys = result["properties"]["ys"]
    zs = get_reference(ys, result)["properties"]["zs"]
    xs = get_reference(zs, result)["properties"]
    assert "required" in result
    assert list(sorted(result["properties"])) == ["id", "ys"]
    assert xs["id"]["description"] == "primary key"


def test_properties__infinite_loop2():
    from alchemyjsonschema import AlsoChildrenWalker, ComfortableDesicion
    target = _makeOne(AlsoChildrenWalker, relation_decision=ComfortableDesicion())
    result = target(X)
    assert "required" in result
    assert list(sorted(result["properties"])) == ["id", "y_id"]

    assert result["properties"]["y_id"] == {"type": "integer", "relation": "ys"}
